@@ -4,7 +4,7 @@
  * Main developers: Bilegsaikhan Naidan, Leonid Boytsov, Yury Malkov, Ben Frederickson, David Novak
  *
  * For the complete list of contributors and further details see:
- * https://github.com/searchivarius/NonMetricSpaceLib
+ * https://github.com/nmslib/nmslib
  *
  * Copyright (c) 2013-2018
  *
@@ -44,6 +44,19 @@
 
 namespace similarity {
 
+
+    typedef float (*EfficientDistFunc)(const float *pVect1, const float *pVect2, size_t &qty, float * TmpRes);
+
+    enum DistFuncType {
+      kDistTypeUnknown = -1,
+      kL2Sqr16Ext = 1,
+      kL2SqrExt = 2,
+      kNormCosine = 3,
+      kNegativeDotProduct = 4,
+      kL1Norm = 5,
+      kLInfNorm = 6
+    };
+
     using std::string;
     using std::vector;
     using std::thread;
@@ -112,11 +125,6 @@ namespace similarity {
                 resultSet1.emplace(curen2.getDistance(), curen2.getMSWNodeHier());
             }
         };
-        /*
-        * Experimental method for selection of neighbors based on local search.
-        * However is peforms worse than the previous method
-        * Note that it works correctly only in a single thread!!!
-        */
 
         template <typename dist_t>
         void getNeighborsByHeuristic2(priority_queue<HnswNodeDistCloser<dist_t>> &resultSet1, const int NN,
@@ -153,15 +161,7 @@ namespace similarity {
                 }
                 if (good)
                     returnlist.push_back(curen);
-                // else
-                //    templist.push(curen);
             }
-            // cout << returnlist.size() <<", looked:"<< h<<" ("<<h/(1.0*full) <<")
-            // missied:"<< (full-h) / (1.0*full)<<endl;
-            /*while (returnlist.size() < NN && templist.size() > 0) {
-            returnlist.push_back(templist.top());
-            templist.pop();
-            }*/
 
             for (HnswNodeDistFarther<dist_t> curen2 : returnlist) {
                 resultSet1.emplace(curen2.getDistance(), curen2.getMSWNodeHier());
@@ -472,11 +472,8 @@ namespace similarity {
         typedef std::vector<HnswNode *> ElementList;
         void baseSearchAlgorithmOld(KNNQuery<dist_t> *query);
         void baseSearchAlgorithmV1Merge(KNNQuery<dist_t> *query);
-        void listPassingModifiedAlgorithm(KNNQuery<dist_t> *query);
-        void SearchL2CustomV1Merge(KNNQuery<dist_t> *query);
-        void SearchL2CustomOld(KNNQuery<dist_t> *query);
-        void SearchCosineNormalizedOld(KNNQuery<dist_t> *query);
-        void SearchCosineNormalizedV1Merge(KNNQuery<dist_t> *query);
+        void SearchOld(KNNQuery<dist_t> *query, bool normalize);
+        void SearchV1Merge(KNNQuery<dist_t> *query, bool normalize);
 
         int getRandomLevel(double revSize)
         {
@@ -485,6 +482,21 @@ namespace similarity {
             return (int)r;
         }
 
+
+        void NormalizeVect(float *v, size_t qty) {
+            float sum = 0;
+            for (int i = 0; i < qty; i++) {
+                sum += v[i] * v[i];
+            }
+            if (sum != 0.0) {
+                sum = 1 / sqrt(sum);
+                for (int i = 0; i < qty; i++) {
+                  v[i] *= sum;
+                }
+            }
+        }
+
+
         void SaveOptimizedIndex(std::ostream& output);
         void LoadOptimizedIndex(std::istream& input);
 
@@ -537,13 +549,13 @@ namespace similarity {
         ElementList ElList_;
 
         int vectorlength_ = 0;
-        int dist_func_type_;
+        DistFuncType dist_func_type_ = kDistTypeUnknown;
         bool iscosine_ = false;
         size_t offsetData_, offsetLevel0_;
         char *data_level0_memory_;
         char **linkLists_;
         size_t memoryPerObject_;
-        float (*fstdistfunc_)(const float *pVect1, const float *pVect2, size_t &qty, float *TmpRes);
+        EfficientDistFunc fstdistfunc_;
 
         enum AlgoType { kOld, kV1Merge, kHybrid };
 
@@ -564,7 +576,8 @@ namespace similarity {
         {
             curV = -1;
             numelements = numelements1;
-            mass = new vl_type[numelements];
+            // we allocate an extra sentinel element to prevent prefetch from accessing out of range memory
+            mass = new vl_type[numelements + 1];
         }
         void reset()
         {
