@@ -4,7 +4,7 @@
  * Main developers: Bilegsaikhan Naidan, Leonid Boytsov, Yury Malkov, Ben Frederickson, David Novak
  *
  * For the complete list of contributors and further details see:
- * https://github.com/searchivarius/NonMetricSpaceLib
+ * https://github.com/nmslib/nmslib
  *
  * Copyright (c) 2013-2018
  *
@@ -26,16 +26,17 @@
 #include <cmath>
 #include <iostream>
 #include <memory>
-// This is only for _mm_prefetch
-#include <mmintrin.h>
 
+#include "portable_prefetch.h"
 #include "portable_simd.h"
 #include "knnquery.h"
 #include "method/hnsw.h"
+#include "method/hnsw_distfunc_opt_impl_inline.h"
 #include "ported_boost_progress.h"
 #include "rangequery.h"
 #include "space.h"
 #include "space/space_lp.h"
+#include "space/space_scalar.h"
 #include "thread_pool.h"
 #include "utils.h"
 
@@ -51,12 +52,6 @@
 #define USE_BITSET_FOR_INDEXING 1
 #define EXTEND_USE_EXTENDED_NEIGHB_AT_CONSTR (0) // 0 is faster build, 1 is faster search on clustered data
 
-#if defined(__GNUC__)
-#define PORTABLE_ALIGN16 __attribute__((aligned(16)))
-#else
-#define PORTABLE_ALIGN16 __declspec(align(16))
-#endif
-
 // For debug purposes we also implemented saving an index to a text file
 #define USE_TEXT_REGULAR_INDEX (false)
 
@@ -68,17 +63,51 @@
 #define FIELD_MAX_M0    "MAX_M0"
 #define CURR_LEVEL      "CURR_LEVEL"
 
+#define EXTRA_MEM_PAD_SIZE 64
+
 namespace similarity {
 
-    // This is the counter to keep the size of neighborhood information (for one node)
+    float
+    NegativeDotProduct(const float *pVect1, const float *pVect2, size_t &qty, float * __restrict TmpRes) {
+        return -ScalarProduct(pVect1, pVect2, qty, TmpRes);
+    }
+
+    /*
+     * Important note: This function is applicable only when both vectors are normalized!
+     */
+    float
+    NormCosine(const float *pVect1, const float *pVect2, size_t &qty, float *__restrict TmpRes) {
+        return std::max(0.0f, 1 - std::max(float(-1), std::min(float(1), ScalarProduct(pVect1, pVect2, qty, TmpRes))));
+    }
+
+    float L1NormWrapper(const float *pVect1, const float *pVect2, size_t &qty, float *) {
+        return L1NormSIMD(pVect1, pVect2, qty);
+    }
+
+    float LInfNormWrapper(const float *pVect1, const float *pVect2, size_t &qty, float *) {
+        return LInfNormSIMD(pVect1, pVect2, qty);
+    }
+
+    EfficientDistFunc getDistFunc(DistFuncType funcType) {
+        switch (funcType) {
+            case kL2Sqr16Ext : return L2Sqr16Ext;
+            case kL2SqrExt   : return L2SqrExt;
+            case kNormCosine : return NormCosine;
+            case kNegativeDotProduct : return NegativeDotProduct;
+            case kL1Norm : return L1NormWrapper;
+            case kLInfNorm : return LInfNormWrapper;
+        }
+
+        return nullptr;
+    }
+
+
+
+// This is the counter to keep the size of neighborhood information (for one node)
     // TODO Can this one overflow? I really doubt
     typedef uint32_t SIZEMASS_TYPE;
 
     using namespace std;
-    /*Functions from hnsw_distfunc_opt.cc:*/
-    float L2SqrSIMDExt(const float *pVect1, const float *pVect2, size_t &qty, float *TmpRes);
-    float L2SqrSIMD16Ext(const float *pVect1, const float *pVect2, size_t &qty, float *TmpRes);
-    float NormScalarProductSIMD(const float *pVect1, const float *pVect2, size_t &qty, float *TmpRes);
 
     template <typename dist_t>
     Hnsw<dist_t>::Hnsw(bool PrintProgress, const Space<dist_t> &space, const ObjectVector &data)
@@ -150,6 +179,7 @@ namespace similarity {
         out.close();
         return;
     }
+
     template <typename dist_t>
     void
     Hnsw<dist_t>::CreateIndex(const AnyParams &IndexParams)
@@ -330,53 +360,65 @@ namespace similarity {
         }
 
         // Selecting custom made functions
-        if (space_.StrDesc().compare("SpaceLp: p = 2 do we have a special implementation for this p? : 1") == 0 &&
-            sizeof(dist_t) == 4) {
-            LOG(LIB_INFO) << "\nThe space is Euclidean";
+        dist_func_type_ = kDistTypeUnknown;
+
+        // Although we removed double, let's keep this check here
+        CHECK(sizeof(dist_t) == 4);
+
+
+        const SpaceLp<dist_t>* pLpSpace = dynamic_cast<const SpaceLp<dist_t>*>(&space_);
+
+        fstdistfunc_ = nullptr;
+        iscosine_ = false;
+        searchMethod_ = 3; // The same for all "optimized" indices
+        if (pLpSpace != nullptr) {
+            if (pLpSpace->getP() == 2) {
+                LOG(LIB_INFO) << "\nThe space is Euclidean";
+                vectorlength_ = ((dataSectionSize - 16) >> 2);
+                LOG(LIB_INFO) << "Vector length=" << vectorlength_;
+                if (vectorlength_ % 16 == 0) {
+                    LOG(LIB_INFO) << "Thus using an optimised function for base 16";
+                    dist_func_type_ = kL2Sqr16Ext;
+                } else {
+                    LOG(LIB_INFO) << "Thus using function with any base";
+                    dist_func_type_ = kL2SqrExt;
+                }
+            } else if (pLpSpace->getP() == 1) {
+                dist_func_type_ = kL1Norm;
+            } else if (pLpSpace->getP() == -1) {
+                dist_func_type_ = kLInfNorm;
+            }
+        } else if (dynamic_cast<const SpaceCosineSimilarity<dist_t>*>(&space_) != nullptr) {
+            LOG(LIB_INFO) << "\nThe vector space is " << space_.StrDesc();
             vectorlength_ = ((dataSectionSize - 16) >> 2);
             LOG(LIB_INFO) << "Vector length=" << vectorlength_;
-            if (vectorlength_ % 16 == 0) {
-                LOG(LIB_INFO) << "Thus using an optimised function for base 16";
-                fstdistfunc_ = L2SqrSIMD16Ext;
-                dist_func_type_ = 1;
-                searchMethod_ = 3;
-            } else {
-                LOG(LIB_INFO) << "Thus using function with any base";
-                fstdistfunc_ = L2SqrSIMDExt;
-                dist_func_type_ = 2;
-                searchMethod_ = 3;
-            }
-        } else if (space_.StrDesc().compare("CosineSimilarity") == 0 && sizeof(dist_t) == 4) {
-            LOG(LIB_INFO) << "\nThe vectorspace is Cosine Similarity";
+            dist_func_type_ = kNormCosine;
+        } else if (dynamic_cast<const SpaceNegativeScalarProduct<dist_t>*>(&space_) != nullptr) {
+            LOG(LIB_INFO) << "\nThe space is " << SPACE_NEGATIVE_SCALAR;
             vectorlength_ = ((dataSectionSize - 16) >> 2);
             LOG(LIB_INFO) << "Vector length=" << vectorlength_;
-            iscosine_ = true;
-            if (vectorlength_ % 4 == 0) {
-                LOG(LIB_INFO) << "Thus using an optimised function for base 4";
-                fstdistfunc_ = NormScalarProductSIMD;
-                dist_func_type_ = 3;
-                searchMethod_ = 4;
-            } else {
-                LOG(LIB_INFO) << "Thus using function with any base";
-                LOG(LIB_INFO) << "Search method 4 is not allowed in this case";
-                fstdistfunc_ = NormScalarProductSIMD;
-                dist_func_type_ = 3;
-                searchMethod_ = 3;
-            }
-        } else {
+            dist_func_type_ = kNegativeDotProduct;
+        }
+
+        fstdistfunc_ = getDistFunc(dist_func_type_);
+        iscosine_ = (dist_func_type_ == kNormCosine);
+
+        if (fstdistfunc_ == nullptr) {
             LOG(LIB_INFO) << "No appropriate custom distance function for " << space_.StrDesc();
-            // if (searchMethod_ != 0 && searchMethod_ != 1)
             searchMethod_ = 0;
             LOG(LIB_INFO) << "searchMethod			  = " << searchMethod_;
             pmgr.CheckUnused();
             return; // No optimized index
         }
+        CHECK(dist_func_type_ != kDistTypeUnknown);
+
         pmgr.CheckUnused();
         LOG(LIB_INFO) << "searchMethod			  = " << searchMethod_;
         memoryPerObject_ = dataSectionSize + friendsSectionSize;
 
         size_t total_memory_allocated = (memoryPerObject_ * ElList_.size());
-        data_level0_memory_ = (char *)malloc(memoryPerObject_ * ElList_.size());
+        // we allocate a few extra bytes to prevent prefetch from accessing out of range memory
+        data_level0_memory_ = (char *)malloc((memoryPerObject_ * ElList_.size()) + EXTRA_MEM_PAD_SIZE);
         CHECK(data_level0_memory_);
 
         offsetLevel0_ = dataSectionSize;
@@ -399,22 +441,14 @@ namespace similarity {
         if (iscosine_) {
             for (long i = 0; i < ElList_.size(); i++) {
                 float *v = (float *)(data_level0_memory_ + (size_t)i * memoryPerObject_ + offsetData_ + 16);
-                float sum = 0;
-                for (int i = 0; i < vectorlength_; i++) {
-                    sum += v[i] * v[i];
-                }
-                if (sum != 0.0) {
-                    sum = 1 / sqrt(sum);
-                    for (int i = 0; i < vectorlength_; i++) {
-                        v[i] *= sum;
-                    }
-                }
+                NormalizeVect(v, vectorlength_);
             };
         }
 
         /////////////////////////////////////////////////////////
         ////////////////////////////////////////////////////////
-        linkLists_ = (char **)malloc(sizeof(void *) * ElList_.size());
+        // we allocate a few extra bytes to prevent prefetch from accessing out of range memory
+        linkLists_ = (char **)malloc((sizeof(void *) * ElList_.size()) + EXTRA_MEM_PAD_SIZE);
         CHECK(linkLists_);
         for (long i = 0; i < ElList_.size(); i++) {
             if (ElList_[i]->level < 1) {
@@ -525,7 +559,7 @@ namespace similarity {
                     int size = neighbor.size();
                     for (int i = 0; i < size; i++) {
                         HnswNode *node = neighbor[i];
-                        _mm_prefetch((char *)(node)->getData(), _MM_HINT_T0);
+                        PREFETCH((char *)(node)->getData(), _MM_HINT_T0);
                     }
                     for (int i = 0; i < size; i++) {
                         currObj = (neighbor[i])->getData();
@@ -634,7 +668,7 @@ namespace similarity {
 
             // calculate distance to each neighbor
             for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
-                _mm_prefetch((char *)(*iter)->getData(), _MM_HINT_T0);
+                PREFETCH((char *)(*iter)->getData(), _MM_HINT_T0);
             }
 
             for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
@@ -690,9 +724,6 @@ namespace similarity {
         bool useOld = searchAlgoType_ == kOld || (searchAlgoType_ == kHybrid && ef_ >= 1000);
         // cout << "Ef = " << ef_ << " use old = " << useOld << endl;
         switch (searchMethod_) {
-        default:
-            throw runtime_error("Invalid searchMethod: " + ConvertToString(searchMethod_));
-            break;
         case 0:
             /// Basic search using Nmslib data structure:
             if (useOld)
@@ -700,24 +731,16 @@ namespace similarity {
             else
                 const_cast<Hnsw *>(this)->baseSearchAlgorithmV1Merge(query);
             break;
-        case 1:
-            /// Experimental search using Nmslib data structure (should not be used):
-            const_cast<Hnsw *>(this)->listPassingModifiedAlgorithm(query);
-            break;
         case 3:
-            /// Basic search using optimized index(cosine+L2)
-            if (useOld)
-                const_cast<Hnsw *>(this)->SearchL2CustomOld(query);
-            else
-                const_cast<Hnsw *>(this)->SearchL2CustomV1Merge(query);
-            break;
         case 4:
-            /// Basic search using optimized index with one-time normalized cosine similarity
-            /// Only for cosine similarity!
+            /// Basic search using optimized index for l2, cosine, negative dot product
             if (useOld)
-                const_cast<Hnsw *>(this)->SearchCosineNormalizedOld(query);
+                const_cast<Hnsw *>(this)->SearchOld(query, iscosine_);
             else
-                const_cast<Hnsw *>(this)->SearchCosineNormalizedV1Merge(query);
+                const_cast<Hnsw *>(this)->SearchV1Merge(query, iscosine_);
+            break;
+        default:
+                throw runtime_error("Invalid searchMethod: " + ConvertToString(searchMethod_));
             break;
         };
     }
@@ -732,16 +755,16 @@ namespace similarity {
 
         unsigned int optimIndexFlag = data_level0_memory_ != nullptr;
 
+        writeBinaryPOD(output, optimIndexFlag);
 
         if (!optimIndexFlag) {
 #if USE_TEXT_REGULAR_INDEX
             SaveRegularIndexText(output);
 #else
-            writeBinaryPOD(output, optimIndexFlag);
+
             SaveRegularIndexBin(output);
 #endif
         } else {
-            writeBinaryPOD(output, optimIndexFlag);
             SaveOptimizedIndex(output);
         }
 
@@ -871,7 +894,7 @@ namespace similarity {
         ReadField(input, FIELD_MAX_M0, maxM0_); lineNum++;
 
         fstdistfunc_ = nullptr;
-        dist_func_type_ = 0;
+        dist_func_type_ = kDistTypeUnknown;
         searchMethod_ = 0;
 
         ElList_.resize(totalElementsStored_);
@@ -927,20 +950,22 @@ namespace similarity {
         readBinaryPOD(input, maxM0_);
 
         fstdistfunc_ = nullptr;
-        dist_func_type_ = 0;
+        dist_func_type_ = kDistTypeUnknown;
         searchMethod_ = 0;
 
         CHECK_MSG(totalElementsStored_ == this->data_.size(),
-             "The number of stored elements " + ConvertToString(totalElementsStored_) + 
-             " doesn't match the number of data points " + ConvertToString(this->data_.size() + 
-             "! Did you forget to re-load data?"))
+             "The number of stored elements " + ConvertToString(totalElementsStored_) +
+             " doesn't match the number of data points " + ConvertToString(this->data_.size()) +
+             "! Did you forget to re-load data?")
 
         ElList_.resize(totalElementsStored_);
         for (unsigned id = 0; id < totalElementsStored_; ++id) {
             ElList_[id] = new HnswNode(this->data_[id], id);
         }
 
-        enterpoint_ = ElList_[enterpointId_];
+        if (!ElList_.empty()) {
+            enterpoint_ = ElList_[enterpointId_];
+        }
 
         for (unsigned id = 0; id < totalElementsStored_; ++id) {
             HnswNode& node = *ElList_[id];
@@ -1015,20 +1040,19 @@ namespace similarity {
 
         LOG(LIB_INFO) << "searchMethod: " << searchMethod_;
 
-        if (dist_func_type_ == 1)
-            fstdistfunc_ = L2SqrSIMD16Ext;
-        else if (dist_func_type_ == 2)
-            fstdistfunc_ = L2SqrSIMDExt;
-        else if (dist_func_type_ == 3)
-            fstdistfunc_ = NormScalarProductSIMD;
+        fstdistfunc_ = getDistFunc(dist_func_type_);
+        iscosine_ = (dist_func_type_ == kNormCosine);
+        CHECK_MSG(fstdistfunc_ != nullptr, "Unknown distance function code: " + ConvertToString(dist_func_type_));
 
         //        LOG(LIB_INFO) << input.tellg();
         LOG(LIB_INFO) << "Total: " << totalElementsStored_ << ", Memory per object: " << memoryPerObject_;
         size_t data_plus_links0_size = memoryPerObject_ * totalElementsStored_;
-        data_level0_memory_ = (char *)malloc(data_plus_links0_size);
+        // we allocate a few extra bytes to prevent prefetch from accessing out of range memory
+        data_level0_memory_ = (char *)malloc(data_plus_links0_size + EXTRA_MEM_PAD_SIZE);
         CHECK(data_level0_memory_);
         input.read(data_level0_memory_, data_plus_links0_size);
-        linkLists_ = (char **)malloc(sizeof(void *) * totalElementsStored_);
+        // we allocate a few extra bytes to prevent prefetch from accessing out of range memory
+        linkLists_ = (char **)malloc( (sizeof(void *) * totalElementsStored_) + EXTRA_MEM_PAD_SIZE);
         CHECK(linkLists_);
 
         data_rearranged_.resize(totalElementsStored_);
@@ -1073,7 +1097,7 @@ namespace similarity {
 
                 const vector<HnswNode *> &neighbor = curNode->getAllFriends(i);
                 for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
-                    _mm_prefetch((char *)(*iter)->getData(), _MM_HINT_T0);
+                    PREFETCH((char *)(*iter)->getData(), _MM_HINT_T0);
                 }
                 for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
                     currObj = (*iter)->getData();
@@ -1120,8 +1144,8 @@ namespace similarity {
             size_t curId;
 
             for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
-                _mm_prefetch((char *)(*iter)->getData(), _MM_HINT_T0);
-                _mm_prefetch((char *)(massVisited + (*iter)->getId()), _MM_HINT_T0);
+                PREFETCH((char *)(*iter)->getData(), _MM_HINT_T0);
+                PREFETCH((char *)(massVisited + (*iter)->getId()), _MM_HINT_T0);
             }
             // calculate distance to each neighbor
             for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
@@ -1171,7 +1195,7 @@ namespace similarity {
 
                 const vector<HnswNode *> &neighbor = curNode->getAllFriends(i);
                 for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
-                    _mm_prefetch((char *)(*iter)->getData(), _MM_HINT_T0);
+                    PREFETCH((char *)(*iter)->getData(), _MM_HINT_T0);
                 }
                 for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
                     currObj = (*iter)->getData();
@@ -1217,10 +1241,10 @@ namespace similarity {
             size_t curId;
 
             for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
-                _mm_prefetch((char *)(*iter)->getData(), _MM_HINT_T0);
+                PREFETCH((char *)(*iter)->getData(), _MM_HINT_T0);
                 IdType curId = (*iter)->getId();
                 CHECK(curId >= 0 && curId < this->data_.size());
-                _mm_prefetch((char *)(massVisited + curId), _MM_HINT_T0);
+                PREFETCH((char *)(massVisited + curId), _MM_HINT_T0);
             }
             // calculate distance to each neighbor
             for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
@@ -1241,7 +1265,7 @@ namespace similarity {
             }
 
             if (itemQty) {
-                _mm_prefetch(const_cast<const char *>(reinterpret_cast<char *>(&itemBuff[0])), _MM_HINT_T0);
+                PREFETCH(const_cast<const char *>(reinterpret_cast<char *>(&itemBuff[0])), _MM_HINT_T0);
                 std::sort(itemBuff.begin(), itemBuff.begin() + itemQty);
 
                 size_t insIndex = 0;
@@ -1274,149 +1298,7 @@ namespace similarity {
 
         visitedlistpool->releaseVisitedList(vl);
     }
-    // Experimental search algorithm
-    template <typename dist_t>
-    void
-    Hnsw<dist_t>::listPassingModifiedAlgorithm(KNNQuery<dist_t> *query)
-    {
-        int efSearchL = 4; // This parameters defines the confidence of searches at level higher than zero
-                           // for zero level it is set to ef
-                           // Getting the visitedlist
-        VisitedList *vl = visitedlistpool->getFreeVisitedList();
-        vl_type *massVisited = vl->mass;
-        vl_type currentV = vl->curV;
 
-        int maxlevel1 = enterpoint_->level;
-
-        const Object *currObj = enterpoint_->getData();
-
-        dist_t d = query->DistanceObjLeft(currObj);
-        dist_t curdist = d;
-        HnswNode *curNode = enterpoint_;
-
-        priority_queue<HnswNodeDistFarther<dist_t>> candidateQueue; // the set of elements which we can use to evaluate
-        priority_queue<HnswNodeDistCloser<dist_t>> closestDistQueue =
-            priority_queue<HnswNodeDistCloser<dist_t>>(); // The set of closest found elements
-        priority_queue<HnswNodeDistCloser<dist_t>> closestDistQueueCpy = priority_queue<HnswNodeDistCloser<dist_t>>();
-
-        HnswNodeDistFarther<dist_t> ev(curdist, curNode);
-        candidateQueue.emplace(curdist, curNode);
-        closestDistQueue.emplace(curdist, curNode);
-
-        massVisited[curNode->getId()] = currentV;
-
-        for (int i = maxlevel1; i > 0; i--) {
-            while (!candidateQueue.empty()) {
-                auto iter = candidateQueue.top();
-                const HnswNodeDistFarther<dist_t> &currEv = iter;
-                // Check condtion to end the search
-                dist_t lowerBound = closestDistQueue.top().getDistance();
-                if (currEv.getDistance() > lowerBound) {
-                    break;
-                }
-
-                HnswNode *initNode = currEv.getMSWNodeHier();
-                candidateQueue.pop();
-
-                const vector<HnswNode *> &neighbor = (initNode)->getAllFriends(i);
-
-                size_t curId;
-
-                for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
-                    _mm_prefetch((char *)(*iter)->getData(), _MM_HINT_T0);
-                    _mm_prefetch((char *)(massVisited + (*iter)->getId()), _MM_HINT_T0);
-                }
-                // calculate distance to each neighbor
-                for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
-                    curId = (*iter)->getId();
-                    if (!(massVisited[curId] == currentV)) {
-                        massVisited[curId] = currentV;
-                        currObj = (*iter)->getData();
-                        d = query->DistanceObjLeft(currObj);
-                        if (closestDistQueue.top().getDistance() > d || closestDistQueue.size() < efSearchL) {
-                            candidateQueue.emplace(d, *iter);
-                            closestDistQueue.emplace(d, *iter);
-                            if (closestDistQueue.size() > efSearchL) {
-                                closestDistQueue.pop();
-                            }
-                        }
-                    }
-                }
-            }
-            // Updating the bitset key:
-            currentV++;
-            vl->curV++; // not to forget updating in the pool
-            if (currentV == 0) {
-                memset(massVisited, 0, ElList_.size() * sizeof(vl_type));
-                currentV++;
-                vl->curV++; // not to forget updating in the pool
-            }
-            candidateQueue = priority_queue<HnswNodeDistFarther<dist_t>>();
-            closestDistQueueCpy = priority_queue<HnswNodeDistCloser<dist_t>>(closestDistQueue);
-            if (i > 1) { // Passing the closest neighbors to layers higher than zero:
-                while (closestDistQueueCpy.size() > 0) {
-                    massVisited[closestDistQueueCpy.top().getMSWNodeHier()->getId()] = currentV;
-                    candidateQueue.emplace(closestDistQueueCpy.top().getDistance(), closestDistQueueCpy.top().getMSWNodeHier());
-                    closestDistQueueCpy.pop();
-                }
-            } else { // Passing the closest neighbors to the 0 zero layer(one has to add also to query):
-                while (closestDistQueueCpy.size() > 0) {
-                    massVisited[closestDistQueueCpy.top().getMSWNodeHier()->getId()] = currentV;
-                    candidateQueue.emplace(closestDistQueueCpy.top().getDistance(), closestDistQueueCpy.top().getMSWNodeHier());
-                    query->CheckAndAddToResult(closestDistQueueCpy.top().getDistance(),
-                                               closestDistQueueCpy.top().getMSWNodeHier()->getData());
-                    closestDistQueueCpy.pop();
-                }
-            }
-        }
-
-        ////////////////////////////////////////////////////////////////////////////////
-        // PHASE TWO OF THE SEARCH
-        // Extraction of the neighborhood to find k nearest neighbors.
-        ////////////////////////////////////////////////////////////////////////////////
-
-        while (!candidateQueue.empty()) {
-            auto iter = candidateQueue.top();
-            const HnswNodeDistFarther<dist_t> &currEv = iter;
-            // Check condtion to end the search
-            dist_t lowerBound = closestDistQueue.top().getDistance();
-            if (currEv.getDistance() > lowerBound) {
-                break;
-            }
-
-            HnswNode *initNode = currEv.getMSWNodeHier();
-            candidateQueue.pop();
-
-            const vector<HnswNode *> &neighbor = (initNode)->getAllFriends(0);
-
-            size_t curId;
-
-            for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
-                _mm_prefetch((char *)(*iter)->getData(), _MM_HINT_T0);
-                _mm_prefetch((char *)(massVisited + (*iter)->getId()), _MM_HINT_T0);
-            }
-            // calculate distance to each neighbor
-            for (auto iter = neighbor.begin(); iter != neighbor.end(); ++iter) {
-                curId = (*iter)->getId();
-                if (!(massVisited[curId] == currentV)) {
-                    massVisited[curId] = currentV;
-                    currObj = (*iter)->getData();
-                    d = query->DistanceObjLeft(currObj);
-                    if (closestDistQueue.top().getDistance() > d || closestDistQueue.size() < ef_) {
-                        {
-                            query->CheckAndAddToResult(d, currObj);
-                            candidateQueue.emplace(d, *iter);
-                            closestDistQueue.emplace(d, *iter);
-                            if (closestDistQueue.size() > ef_) {
-                                closestDistQueue.pop();
-                            }
-                        }
-                    }
-                }
-            }
-        }
-        visitedlistpool->releaseVisitedList(vl);
-    }
 
     template class Hnsw<float>;
     template class Hnsw<int>;
