@@ -25,6 +25,9 @@ MEM_TEST_QUERY_QTY=200
 MEM_GROW_COEFF=1.5 # This is a bit adhoc but seems to work in practice
 MEM_TEST_DATA_DIM=4
 
+RTOL=1e-7
+ATOL=1e-7
+
 def get_exact_cosine(row, data, N=10):
     scores = data.dot(row) / np.linalg.norm(data, axis=-1)
     best = np.argpartition(scores, -N)[-N:]
@@ -63,7 +66,48 @@ class TestCaseBase(unittest.TestCase):
         comp_resort_dists.append(ids2dist[one_id])
 
       npt.assert_allclose(orig,
-                          (comp_resort_ids, comp_resort_dists))
+                          (comp_resort_ids, comp_resort_dists), 
+                           rtol=RTOL, atol=ATOL)
+
+class TestSparseZeroVect(TestCaseBase):
+    def testSparseCosine1(self):
+        X = np.array([[4., 2., 3., 1., 0., 0., 0., 0., 0.],
+                      [2., 1., 0., 0., 3., 0., 1., 2., 1.],
+                      [4., 2., 0., 0., 3., 1., 0., 0., 0.],
+                      [0., 0., 0., 0., 0., 0., 0., 0., 0.]], dtype=np.float32)
+
+        q = np.array([[0., 0., 0., 0., 0., 0., 0., 0., 0.]], dtype=np.float32)
+
+        index = nmslib.init(method='hnsw', space='cosinesimil')
+        index.addDataPointBatch(X)
+        index.createIndex()
+        nns, dists = index.knnQueryBatch(q, k=3)[0]
+        self.assertEqual(len(nns), 3)
+        self.assertEqual(len(dists), 3)
+        npt.assert_allclose(dists, [1, 1, 1], rtol=RTOL, atol=ATOL)
+        self.assertEqual(index.getDistance(3, 3), 1)
+        # The last element is all zeros, so any distance to it must be equal to one
+        for k in range(len(X)):
+            self.assertEqual(index.getDistance(k, len(X)-1), 1)
+
+
+    def testSparseNoFailOnEmpty(self):
+        # These are three sparse vectors, the last one is empty
+        data = [ [(0,1)], [(1,1)], [] ]
+        qty = len(data)
+
+        for space in ['cosinesimil_sparse', 'cosinesimil_sparse_fast', 'cosinesimil_sparse_bin_fast', 'l2_sparse', 'l1_sparse']:
+            index = nmslib.init(method='hnsw', space='cosinesimil_sparse', data_type=nmslib.DataType.SPARSE_VECTOR) 
+            index.addDataPointBatch(data)
+            self.assertEqual(len(index), qty)
+        
+            index.createIndex({'M': 30, 'indexThreadQty': 4, 'efConstruction': 100, 'post' : 0}) 
+  
+            if space.startswith('cosine'):
+                # The last element is all zeros, so any distance to it must be equal to one
+                for k in range(qty):
+                    self.assertEqual(index.getDistance(k, qty - 1), 1)
+        
 
 class DenseIndexTestMixin(object):
     def _get_index(self, space='cosinesimil'):
@@ -272,7 +316,7 @@ class BallTreeTestCase(TestCaseBase, DenseIndexTestMixin):
         return nmslib.init(method='vptree', space=space)
 
     def testReloadIndex(self):
-        return NotImplemented
+        pass
 
 
 class StringTestCase(TestCaseBase):
@@ -312,7 +356,7 @@ class SparseTestCase(TestCaseBase):
 
         ids, distances = index.knnQuery([(1, 2.), (2, 3.)])
         self.assertEqual(ids[0], 0)
-        npt.assert_allclose(distances[0], 0)
+        npt.assert_allclose(distances[0], 0, rtol=RTOL, atol=ATOL)
 
         self.assertEqual(len(index), 4)
         self.assertEqual(index[3], [(3, 1.0)])
@@ -471,6 +515,21 @@ class MemoryLeak2TestCase(TestCaseBase):
         print('Fail qty %d out of %d' % (fail_qty, test_qty))
         self.assertTrue(fail_qty < MEM_TEST_ITER2 * MEM_TEST_CRIT_FAIL_RATE)
 
+class TestLoadEmptyIndex(TestCaseBase):
+    # A regression test for issue #452
+    def testRegression452(self):
+        temp_dir = tempfile.mkdtemp()
+        temp_file_pref = os.path.join(temp_dir, 'index')
+
+        for skip_optim in [0, 1]:
+            print(skip_optim)
+            index_params = {'skip_optimized_index' : skip_optim}
+            index = nmslib.init(method='hnsw', space='l2')
+            index.createIndex()
+            index.saveIndex(temp_file_pref, True)
+            index.loadIndex(temp_file_pref, True)
+
+        shutil.rmtree(temp_dir)
 
 
 class GlobalTestCase(TestCaseBase):
